/*
* Copyright © 2010 Martin Gräßlin <kde@martin-graesslin.com>
* Copyright © 2008 Dennis Kasprzyk <onestone@opencompositing.org>
* Copyright © 2006 Novell, Inc.
* Copyright © 2006 Volker Krause <vkrause@kde.org>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
* MA 02110-1301, USA.
*
* Author: David Reveman <davidr@novell.com>
*/
#include "aurorae.h"
#include "client.h"
#include "utils.h"
// Aurorae
#include "lib/auroraetheme.h"
// Qt
// KDE
#include <KDE/KDebug>
#include <KDE/KConfig>
#include <KDE/KConfigGroup>
#include <KDE/KWindowSystem>
// X11
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/extensions/Xcomposite.h>
#include <fixx11h.h>


NETRootInfo *Aurorae::AuroraeDecoration::m_rootInfo;

namespace Aurorae
{

AuroraeDecoration::AuroraeDecoration()
    : KApplication()
    , m_theme(new AuroraeTheme(this))
    , m_activeId(0)
{
    m_rootInfo = new NETRootInfo(QX11Info::display(), 0);
    m_theme->setCompositingActive(true);
    XSetWindowAttributes attr;
    Atoms::init();
    attr.override_redirect = True;
    m_compositeWindow = XCreateWindow (QX11Info::display(),
                    QX11Info::appRootWindow(),
                    -ROOT_OFF_X, -ROOT_OFF_Y, 1, 1, 0,
                    CopyFromParent,
                    CopyFromParent,
                    CopyFromParent,
                    CWOverrideRedirect, &attr);

    long data = 1;
    XChangeProperty (QX11Info::display(), m_compositeWindow, Atoms::enlightmentDesktop,
              XA_CARDINAL, 32, PropModeReplace, (unsigned char *) &data, 1);

    XCompositeRedirectSubwindows(QX11Info::display(), m_compositeWindow,
                  CompositeRedirectManual);

    XMapWindow(QX11Info::display(), m_compositeWindow);
}

AuroraeDecoration::~AuroraeDecoration()
{
    QMap <WId, Client *>::ConstIterator it;

    for (it = m_clients.begin(); it != m_clients.end (); ++it) {
        delete (*it);
    }

    XDestroyWindow (QX11Info::display (), m_compositeWindow);
}

bool AuroraeDecoration::enableDecorations()
{
    KConfig conf("auroraerc");
    KConfigGroup group(&conf, "Engine");

    const QString themeName = group.readEntry("ThemeName", "example-deco");
    KConfig config("aurorae/themes/" + themeName + '/' + themeName + "rc", KConfig::FullConfig, "data");
    KConfigGroup themeGroup(&conf, themeName);
    m_theme->loadTheme(themeName, config);
    m_theme->setBorderSize((KDecorationDefines::BorderSize)themeGroup.readEntry<int>("BorderSize", KDecorationDefines::BorderNormal));
    m_theme->setButtonSize((KDecorationDefines::BorderSize)themeGroup.readEntry<int>("ButtonSize", KDecorationDefines::BorderNormal));
    m_theme->setShowTooltips(true);

    m_activeId = KWindowSystem::activeWindow();

    connect(KWindowSystem::self(), SIGNAL(windowAdded(WId)),
         SLOT(handleWindowAdded(WId)));
    connect(KWindowSystem::self(), SIGNAL(windowRemoved(WId)),
         SLOT(handleWindowRemoved(WId)));
    connect(KWindowSystem::self(), SIGNAL(activeWindowChanged(WId)),
         SLOT(handleActiveWindowChanged(WId)));
    connect(KWindowSystem::self(), SIGNAL(windowChanged(WId, const unsigned long *)),
         SLOT(handleWindowChanged (WId, const unsigned long *)));

    foreach (WId id, KWindowSystem::windows()) {
        handleWindowAdded(id);
    }

    // select for client messages
    XSelectInput(QX11Info::display (), QX11Info::appRootWindow (),
          StructureNotifyMask | PropertyChangeMask);

    return true;
}

void AuroraeDecoration::handleActiveWindowChanged(WId id)
{
    if (id != m_activeId) {
        Client *newActiveWindow = 0;
        Client *oldActiveWindow = 0;

        if (m_clients.contains(id)) {
            newActiveWindow = m_clients[id];
        }

        if (m_clients.contains(m_activeId)) {
            oldActiveWindow = m_clients[m_activeId];
        }

        m_activeId = id;

        if (oldActiveWindow) {
            oldActiveWindow->handleActiveChange(false);
        }

        if (newActiveWindow) {
            newActiveWindow->handleActiveChange(true);
        }
    }
}

void AuroraeDecoration::handleWindowAdded(WId id)
{
    QMap<WId, Client *>::ConstIterator it;
    Client *client = 0;
    WId frame = 0;
    WId oframe = 0, iframe = 0;
    QWidgetList widgets;

    /* avoid adding any of our own top level windows */
    foreach (QWidget *widget, QApplication::topLevelWidgets ()) {
        if (widget->winId() == id) {
            return;
        }
    }

    KWindowInfo wInfo = KWindowSystem::windowInfo(id, NET::WMGeometry);

    if (!wInfo.valid()) {
        return;
    }

    const QRect geometry = wInfo.geometry ();

    Aurorae::readWindowProperty(id, Atoms::netInputFrameWindow, (long *) &iframe);
    Aurorae::readWindowProperty(id, Atoms::netOutputFrameWindow, (long *) &oframe);

    wInfo = KWindowSystem::windowInfo(id, NET::WMWindowType, 0);

    switch (wInfo.windowType(~0)) {
    case NET::Normal:
    case NET::Dialog:
    case NET::Toolbar:
    case NET::Menu:
    case NET::Utility:
    case NET::Splash:
    case NET::Unknown:
        /* decorate these window types */
        break;
    default:
        return;
    }

    if (iframe) {
        frame = iframe;
    } else {
        frame = oframe;
    }

    XSelectInput (QX11Info::display(), id,
          StructureNotifyMask | PropertyChangeMask);

    if (frame) {
        XWindowAttributes attr;
        XGetWindowAttributes (QX11Info::display (), frame, &attr);
        if (!m_clients.contains(id)) {
            kDebug() << "Client will be created";
            client = new Client(m_theme, m_compositeWindow, id, frame, geometry);

            m_clients.insert(id, client);
            m_frames.insert(frame, client);
        } else {
            client = m_clients[id];
            m_frames.remove(client->frameId());
            m_frames.insert(frame, client);

            client->updateFrame(frame);
        }
    } else {
        if (m_clients.contains(id)) {
            client = m_clients[id];
        }

        if (client) {
            m_clients.remove(client->windowId());
            m_frames.remove(client->frameId());

            delete client;
        }
    }
}

void AuroraeDecoration::handleWindowChanged(WId id, const long unsigned int* properties)
{
    Client *client;

    if (!m_clients.contains(id)) {
        return;
    }

    client = m_clients[id];

    if (properties[0] & NET::WMName) {
        client->updateName();
    }
    if (properties[0] & NET::WMVisibleName) {
        client->updateName();
    }
    if (properties[0] & NET::WMState) {
        client->updateState();
    }
    if (properties[0] & NET::WMIcon) {
        client->updateIcons();
    }
    if (properties[0] & NET::WMGeometry) {
        client->updateWindowGeometry();
    }
}

void AuroraeDecoration::handleWindowRemoved(WId id)
{
    Client *window = 0;

    if (m_clients.contains(id)) {
        window = m_clients[id];
    } else if (m_frames.contains(id)) {
        window = m_frames[id];
    }

    if (window) {
        m_clients.remove(window->windowId());
        m_frames.remove(window->frameId());
        delete window;
    }
}

} // namespace
