/*
* Copyright © 2010 Martin Gräßlin <kde@martin-graesslin.com>
* Copyright © 2009 Dennis Kasprzyk <onestone@compiz-fusion.org>
* Copyright © 2006 Novell, Inc.
* Copyright © 2006 Volker Krause <vkrause@kde.org>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
* MA 02110-1301, USA.
*
* Author: David Reveman <davidr@novell.com>
*/
#include "client.h"
#include "utils.h"
// Aurorae
#include "lib/auroraescene.h"
#include "lib/auroraetheme.h"
// Compiz
#include <decoration.h>
// Qt
// KDE
#include <KDE/KDebug>
#include <KDE/KWindowSystem>
// X11
#include <X11/Xlib.h>
#include <X11/Xatom.h>

namespace Aurorae
{

Client::Client(Aurorae::AuroraeTheme* theme, WId parentId, WId clientId, WId frameId, const QRect& geometry)
    : QGraphicsView()
    , m_theme(theme)
    , m_parentId(parentId)
    , m_clientId(clientId)
    , m_frameId(frameId)
    , m_geometry(geometry)
    , m_state(0)
{
    m_scene = new AuroraeScene(m_theme, m_theme->defaultButtonsLeft(), m_theme->defaultButtonsRight(), false, this);
    setScene(m_scene);
    init();
}

void Client::init()
{
    setAttribute(Qt::WA_PaintOnScreen);
    setAttribute(Qt::WA_NoSystemBackground);
    setAttribute(Qt::WA_TranslucentBackground);
    setFrameShape(QFrame::NoFrame);
    QPalette pal = palette();
    pal.setColor(backgroundRole(), Qt::transparent);
    setPalette(pal);
    setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    KWindowInfo wInfo = KWindowSystem::windowInfo(m_clientId, NET::WMState |
                               NET::WMVisibleName, 0);

    m_state = wInfo.state();

    m_scene->addTab(QString());
    updateName();
    updateIcons();
    updateFrame(m_frameId);
    m_scene->setActive(false, false);
    m_scene->setAllDesktops(m_state & NET::Sticky);
    m_scene->setMaximizeMode(m_state & NET::Max ? KDecorationDefines::MaximizeFull : KDecorationDefines::MaximizeRestore);
    m_scene->setShade(m_state & NET::Shaded);
    m_scene->setKeepAbove(m_state & NET::KeepAbove);
    m_scene->setKeepBelow(m_state & NET::KeepBelow);
    m_theme->setShowTooltips(true);
    createDecoration();

    connect(m_scene, SIGNAL(closeWindow()), SLOT(closeWindow()));
    connect(m_scene, SIGNAL(menuDblClicked()), SLOT(closeWindow()));
    connect(m_scene, SIGNAL(menuClicked()), SLOT(menuClicked()));
    connect(m_scene, SIGNAL(minimizeWindow()), SLOT(minimizeWindow()));
    connect(m_scene, SIGNAL(maximize(Qt::MouseButtons)), SLOT(maximize(Qt::MouseButtons)));
    connect(m_scene, SIGNAL(titleDoubleClicked()), SLOT(titleDoubleClicked()));
    connect(m_scene, SIGNAL(toggleKeepAbove()), SLOT(toggleKeepAbove()));
    connect(m_scene, SIGNAL(toggleKeepBelow()), SLOT(toggleKeepBelow()));
    connect(m_scene, SIGNAL(toggleOnAllDesktops()), SLOT(toggleOnAllDesktops()));
    connect(m_scene, SIGNAL(toggleShade()), SLOT(toggleShade()));
}

void Client::createDecoration()
{
    int paddingLeft, paddingTop, paddingRight, paddingBottom;
    m_theme->padding(paddingLeft, paddingTop, paddingRight, paddingBottom);
    if (m_frameId) {
        XReparentWindow (QX11Info::display(), winId(), m_frameId, 0, 0);
        XMoveWindow (QX11Info::display(), winId(), -paddingLeft, -paddingTop);
    } else {
        XReparentWindow (QX11Info::display(), winId(), m_parentId, 0, 0);
    }
    updateWindowGeometry();
}

WId Client::frameId() const
{
    return m_frameId;
}

void Client::handleActiveChange(bool active)
{
    m_scene->setActive(active);
}

void Client::updateFrame(WId id)
{
    m_frameId = id;

    XSelectInput(QX11Info::display(), m_frameId,
          StructureNotifyMask | PropertyChangeMask |
          ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
          EnterWindowMask | LeaveWindowMask);
}

void Client::updateIcons()
{
    QPixmap icon = KWindowSystem::icon(m_clientId, 32, 32, true,
                                 KWindowSystem::NETWM | KWindowSystem::WMHints);

    if (icon.isNull ()) {
        icon = KWindowSystem::icon(m_clientId, 32, 32, true,
                     KWindowSystem::ClassHint | KWindowSystem::XApp);
    }
    m_scene->setIcon(QIcon(icon));
}

void Client::updateName()
{
    KWindowInfo wInfo = KWindowSystem::windowInfo(m_clientId, NET::WMVisibleName, 0);
    m_scene->setCaption(wInfo.visibleName());
}

void Client::updateState()
{
    KWindowInfo wInfo = KWindowSystem::windowInfo(m_clientId, NET::WMState, 0);

    unsigned long newState = wInfo.state ();
    unsigned long stateChange = m_state ^ newState;

    m_state = newState;

    if (stateChange & NET::Max) {
        if (m_state & NET::MaxHoriz && m_state & NET::MaxVert) {
            m_scene->setMaximizeMode(KDecorationDefines::MaximizeFull);
        } else {
            m_scene->setMaximizeMode(KDecorationDefines::MaximizeRestore);
        }
        resizeDecoration(false);
    }

    if (stateChange & NET::KeepAbove && !(m_state & NET::KeepAbove)) {
        m_scene->setKeepAbove(m_state & NET::KeepAbove);
    }
    if (stateChange & NET::KeepBelow && !(m_state & NET::KeepBelow)) {
        m_scene->setKeepBelow(m_state & NET::KeepBelow);
    }
    if (stateChange & NET::KeepAbove && m_state & NET::KeepAbove) {
        m_scene->setKeepAbove(m_state & NET::KeepAbove);
    }
    if (stateChange & NET::KeepBelow && m_state & NET::KeepBelow)
        m_scene->setKeepBelow(m_state & NET::KeepBelow);
    if (stateChange & NET::Shaded) {
        m_scene->setShade(m_state & NET::Shaded);
    }
    if (stateChange & NET::Sticky) {
        m_scene->setAllDesktops(m_state & NET::Sticky);
    }
}

void Client::updateWindowGeometry()
{
    KWindowInfo wInfo = KWindowSystem::windowInfo(m_clientId, NET::WMGeometry);

    if (!wInfo.valid()) {
        kDebug() << "Window Info not valid";
        return;
    }

    const QRect geometry = wInfo.geometry ();
    int w, h;

    int borderLeft, borderTop, borderRight, borderBottom;
    m_theme->borders(borderLeft, borderTop, borderRight, borderBottom,
                     false/*m_scene->maximizeMode() == KDecorationDefines::MaximizeFull*/);
    int paddingLeft, paddingTop, paddingRight, paddingBottom;
    m_theme->padding(paddingLeft, paddingTop, paddingRight, paddingBottom);
    w = geometry.width() + borderLeft + borderRight + paddingLeft + paddingRight;
    h = geometry.height() + borderTop + borderBottom + paddingTop + paddingBottom;

    m_geometry = QRect(geometry.x(), geometry.y(), w, h);
    resizeDecoration();
}

WId Client::windowId() const
{
    return m_clientId;
}

void Client::resizeDecoration(bool force)
{
    kDebug() << "Resize decoration";
    m_scene->setSceneRect(QRectF(0, 0, m_geometry.width(), m_geometry.height()));
    m_scene->updateLayout();
    setGeometry(QRect(0, 0, m_geometry.width(), m_geometry.height()));
    updateProperty();
    show();
}

void Client::updateProperty()
{
    Atom        atom = Atoms::netWindowDecor;
    decor_extents_t maxExtents, normExtents;
    long        data[256];
    unsigned int    saveState;

    m_theme->borders (maxExtents.left, maxExtents.top,
             maxExtents.right, maxExtents.bottom, true);
    m_theme->borders (normExtents.left, normExtents.top,
             normExtents.right, normExtents.bottom, false);

    int paddingLeft, paddingTop, paddingRight, paddingBottom;
    m_theme->padding(paddingLeft, paddingTop, paddingRight, paddingBottom);
    normExtents.left = normExtents.left +paddingLeft;
    normExtents.top = normExtents.top +paddingTop;
    normExtents.right = normExtents.right + paddingRight;
    normExtents.bottom = normExtents.bottom + paddingBottom;

    decor_gen_window_property (data, &normExtents, &maxExtents, 1, 0);

    XChangeProperty (QX11Info::display(), m_clientId, atom,
             XA_INTEGER,
             32, PropModeReplace, (unsigned char *) data,
             12);
}

void Client::closeWindow()
{
    AuroraeDecoration::rootInfo()->closeWindowRequest(m_clientId);
}

void Client::maximize(Qt::MouseButtons buttons)
{
    int mode;
    if (buttons.testFlag(Qt::RightButton)) {
        mode = KDecorationDefines::MaximizeVertical;
    }
    if (buttons.testFlag(Qt::MidButton)) {
        mode = KDecorationDefines::MaximizeHorizontal;
    }
    if (buttons.testFlag(Qt::LeftButton)) {
        mode = KDecorationDefines::MaximizeHorizontal | KDecorationDefines::MaximizeVertical;
    }
    if (m_state & NET::MaxHoriz) {
        mode &= ~KDecorationDefines::MaximizeHorizontal;
    }
    if (m_state & NET::MaxVert) {
        mode &= ~KDecorationDefines::MaximizeVertical;
    }
    KWindowSystem::setState (m_clientId,
            ((mode & KDecorationDefines::MaximizeVertical) ? NET::MaxVert : 0) |
            ((mode & KDecorationDefines::MaximizeHorizontal) ? NET::MaxHoriz : 0));
    KWindowSystem::clearState (m_clientId,
              ((mode & KDecorationDefines::MaximizeVertical) ? 0 : NET::MaxVert) |
              ((mode & KDecorationDefines::MaximizeHorizontal) ? 0 : NET::MaxHoriz));
}

void Client::menuClicked()
{
    // TODO: show menu
}

void Client::minimizeWindow()
{
    KWindowSystem::minimizeWindow(m_clientId, false);
}

void Client::titleDoubleClicked()
{
    // TODO: different options - compiz specific
    maximize(Qt::LeftButton);
}

void Client::toggleKeepAbove()
{
    if (m_state & NET::KeepAbove) {
        KWindowSystem::clearState(m_clientId, NET::KeepAbove);
    } else {
        KWindowSystem::setState(m_clientId, NET::KeepAbove);
        KWindowSystem::clearState(m_clientId, NET::KeepBelow);
    }
}

void Client::toggleKeepBelow()
{
    if (m_state & NET::KeepBelow) {
        KWindowSystem::clearState(m_clientId, NET::KeepBelow);
    } else {
        KWindowSystem::setState(m_clientId, NET::KeepBelow);
        KWindowSystem::clearState(m_clientId, NET::KeepAbove);
    }
}

void Client::toggleOnAllDesktops()
{
    if (m_state & NET::Sticky) {
        KWindowSystem::clearState(m_clientId, NET::Sticky);
    } else {
        KWindowSystem::setState(m_clientId, NET::Sticky);
    }
}

void Client::toggleShade()
{
    if (m_state & NET::Shaded) {
        KWindowSystem::clearState(m_clientId, NET::Shaded);
    } else {
        KWindowSystem::setState(m_clientId, NET::Shaded);
    }
}

} // namespace
